"""
Load and validate a data dictionary for use with autoparser.
"""

from __future__ import annotations

import argparse
from pathlib import Path

import pandas as pd
import pandera.pandas as pa
from pandera.typing.pandas import DataFrame

from .config.config import get_config
from .data_dict_schema import DataDictionaryEntry, DataDictionaryProcessed
from .util import DEFAULT_CONFIG, read_data


class DictReader:
    """
    Class for reading in and converting data dictionaries provided by users into a
    format usable by autoparser.

    Validates the final data dictionary against the schema defined in
    `adtl.autoparser.data_dict_schema.DataDictionaryProcessed`.

    Parameters
    ----------
    data_dict
        Path to a CSV, XLSX or parquet file, or a DataFrame, containing the data dictionary.

    config
        The path to the configuration file to use if not using the default configuration

    """

    def __init__(
        self,
        data_dict: pd.DataFrame | str,
    ):
        self.config = get_config()

        self.data_dict = read_data(data_dict, "Data Dictionary")

    def _parse_choices(self, choice_col: pd.Series) -> pd.Series:
        sep = self.config.choice_delimiter
        link = self.config.choice_delimiter_map

        def parse(x):
            if not isinstance(x, str):
                return None
            options = {}
            for i in x.split(sep):
                try:
                    k, v = i.split(link)
                except ValueError:
                    return None
                k = k.strip()
                v = v.strip()
                options[k] = v
            return options if options else None

        return choice_col.apply(lambda x: parse(x))

    def _process_dict(self, dd):
        if "common_values" in dd.columns:

            def _lower_string(x):
                if isinstance(x, str):
                    return list(
                        {
                            y.lower().strip()
                            for y in x.split(self.config.choice_delimiter)
                        }
                    )

            dd["common_values"] = dd["common_values"].apply(_lower_string)

        elif "choices" in dd.columns:
            dd.loc[:, "choices"] = self._parse_choices(dd["choices"])

        return dd

    def validate_dictionary(
        self, save: bool | str = True
    ) -> DataFrame[DataDictionaryProcessed]:
        try:
            # check if the data dictionary is already in the correct format
            # e.g. if loaded from a parquet file
            DataDictionaryProcessed.validate(self.data_dict)
            return self.data_dict
        except pa.errors.SchemaError:
            pass

        # Assumes an unformatted data dictionary, so processes it
        try:
            dd = self.data_dict
            column_mappings = {
                v: k for k, v in self.config.column_mappings.model_dump().items()
            }
            dd.rename(columns=column_mappings, inplace=True)

            dd = dd.loc[
                :,
                dd.columns.isin(list(DataDictionaryProcessed.to_schema().columns)),
            ]

            DataDictionaryEntry.validate(dd, lazy=True)
        except pa.errors.SchemaErrors as exc:
            failure_cases = exc.failure_cases
            failure_cases["field_name"] = exc.failure_cases["index"].map(
                self.data_dict["source_field"]
            )

            with pd.option_context("display.max_rows", None):
                # Order the failure cases by the index of the first occurrence of each failure case
                # and then group by failure case to make clearer
                fc = failure_cases.assign(
                    failure_case=pd.Categorical(
                        exc.failure_cases["failure_case"],
                        categories=exc.failure_cases.groupby("failure_case")["index"]
                        .min()
                        .sort_values()
                        .index,
                        ordered=True,
                    )
                ).sort_values(["failure_case", "index"])

                raise pa.errors.SchemaError(
                    schema=exc.schema,
                    data=exc.data,
                    message=(
                        f"Data dictionary validation failed with {len(exc.failure_cases)} error(s). See below for details.  \n"
                        + f"{fc}"
                    ),
                ) from None

        self.data_dict = self._process_dict(dd)

        try:
            DataDictionaryProcessed.validate(self.data_dict, lazy=True)
        except pa.errors.SchemaErrors as exc:
            raise pa.errors.SchemaError(
                schema=exc.schema,
                data=exc.data,
                message=(
                    f"Processed data dictionary validation failed with {len(exc.failure_cases)} error(s). See below for details.  \n"
                    + f"{exc.failure_cases}"
                ),
            ) from None  # pragma: no cover

        if save:
            self.save_formatted_dictionary(name=save if isinstance(save, str) else None)

        return self.data_dict

    def save_formatted_dictionary(self, name: str | Path | None = None) -> None:
        """
        Save the formatted data dictionary to a parquet file.

        The file will be saved in the same directory as the input data dictionary,
        either with the name provided or as 'formatted_data_dict.parquet'
        """
        if name:
            output_path = Path(name)
            if not output_path.suffix == "parquet":
                output_path = output_path.with_suffix(".parquet")
        else:
            output_path = Path("formatted_data_dict.parquet")

        self.data_dict.to_parquet(output_path, index=False, engine="fastparquet")
        print(f"Formatted data dictionary saved to '{output_path}'")


def format_dict(data_dict: pd.DataFrame | str, save=False) -> pd.DataFrame:
    """
    Formats a pre-existing data dictionary to use with autoparser, or checks one
    that is already pre-formatted.

    Parameters
    ----------
    data_dict
        Path to a CSV, XLSX or parquet file, or a DataFrame, containing the data dictionary.
    config
        Path to the configuration file to use if not using the default configuration
    save
        If True, saves the formatted data dictionary to a parquet file in the same directory as the
        input data dictionary, as 'formatted_data_dictionary.parquet'.

    Returns
    -------
    pd.DataFrame
        Data dictionary containing field names, field types, and common values.
    """

    dr = DictReader(data_dict)
    dd = dr.validate_dictionary(save=save)

    return dd


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Check the formatting of your data dictionary and save a version in the correct format.",
        prog="autoparser format-dict",
    )
    parser.add_argument("data_dict", help="Data to create dictionary from")

    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )

    parser.add_argument("-o", "--output", help="new output data dictionary name")

    args = parser.parse_args(argv)

    dict_reader = DictReader(data_dict=args.data_dict)

    dict_reader.validate_dictionary(save=args.output if args.output else True)


if __name__ == "__main__":
    main()  # pragma: no cover



from __future__ import annotations

import copy
import csv
import hashlib
import io
import itertools
import json
import logging
import re
import uuid
import warnings
from collections import Counter, defaultdict
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import Any, Iterable, Literal, Union

import fastjsonschema
import pandas as pd
import requests
import tomli
from joblib import Parallel, delayed
from more_itertools import unique_everseen
from tqdm.auto import tqdm

import adtl.util as util
from adtl.get_value import get_value, parse_if

SUPPORTED_FORMATS = {"json": json.load, "toml": tomli.load}
DEFAULT_DATE_FORMAT = "%Y-%m-%d"

StrDict = dict[str, Any]

logger = logging.getLogger(__name__)


def expand_refs(spec_fragment: StrDict, defs: StrDict) -> Union[StrDict, list[StrDict]]:
    "Expand all references (ref) with definitions (defs)"

    if spec_fragment == {}:
        return {}
    if isinstance(spec_fragment, dict):
        if "ref" in spec_fragment:
            reference_expanded = defs[spec_fragment["ref"]]
            del spec_fragment["ref"]
            spec_fragment = {**reference_expanded, **spec_fragment}
        return {k: expand_refs(spec_fragment[k], defs) for k in spec_fragment}
    elif isinstance(spec_fragment, list):
        return [expand_refs(m, defs) for m in spec_fragment]
    else:
        return spec_fragment


def expand_for(spec: list[StrDict]) -> list[StrDict]:
    "Expands for expressions in oneToMany table blocks"

    out = []

    def replace_val(
        item: Union[str, float, dict[str, Any]], replace: dict[str, Any]
    ) -> dict[str, Any]:
        block = {}
        if isinstance(item, str):
            return item.format(**replace)
        elif isinstance(item, (float, int)):
            return item
        for k, v in item.items():
            if not isinstance(k, str):
                block[k] = v
                continue
            rk = k.format(**replace)
            if isinstance(v, dict):
                block[rk] = replace_val(v, replace)
            elif isinstance(v, str):
                block[rk] = v.format(**replace)
            elif isinstance(v, list):
                block[rk] = [replace_val(it, replace) for it in v]
            else:
                block[rk] = v
        return block

    for match in spec:
        if "for" not in match:
            out.append(match)
            continue
        for_expr = match.pop("for")
        if not isinstance(for_expr, dict):
            raise ValueError(
                f"for expression {for_expr!r} is not a dictionary of variables to list "
                "of values or a range"
            )

        # Expand ranges when available
        for var in for_expr:
            if (
                "range" in for_expr[var]
                and isinstance(for_expr[var]["range"], list)
                and len(for_expr[var]["range"]) == 2
                and isinstance(for_expr[var]["range"][0], int)
                and isinstance(for_expr[var]["range"][1], int)
                and for_expr[var]["range"][1] > for_expr[var]["range"][0]
            ):
                start, end = for_expr[var]["range"]
                for_expr[var] = range(start, end + 1)  # add one to include end in list
            elif isinstance(for_expr[var], list):
                pass
            else:
                raise ValueError(
                    f"for expression {for_expr!r} can only have lists or ranges for "
                    "variables"
                )
        loop_vars = sorted(for_expr.keys())
        loop_assignments = [
            dict(zip(loop_vars, vals))
            for vals in itertools.product(*(for_expr[var] for var in loop_vars))
        ]
        for replacement in loop_assignments:
            out.append(replace_val(match, replacement))
    return out


def remove_null_keys(d: dict[str, Any]) -> dict[str, Any]:
    "Removes keys which map to null - but not empty strings or 'unknown' etc types"
    return {k: v for k, v in d.items() if v is not None}


def get_date_fields(schema: dict[str, Any]) -> list[str]:
    "Returns list of date fields from schema"
    fields = [
        field
        for field in schema["properties"]
        if field == "date" or "date_" in field or "_date" in field
    ]
    format_date_fields = [
        field
        for field in schema["properties"]
        if schema["properties"][field].get("format") == "date"
    ]
    return sorted(set(fields + format_date_fields))


def make_fields_optional(
    schema: dict[str, Any], optional_fields: list[str]
) -> dict[str, Any]:
    "Returns JSON schema with required fields modified to drop optional fields"
    if optional_fields is None:
        return schema
    _schema = copy.deepcopy(schema)
    _schema["required"] = sorted(set(schema["required"]) - set(optional_fields))
    for opt in ["oneOf", "anyOf"]:
        if opt in _schema:
            if any("required" in _schema[opt][x] for x in range(len(_schema[opt]))):
                for x in range(len(_schema[opt])):
                    _schema[opt][x]["required"] = list(
                        set(_schema[opt][x]["required"]) - set(optional_fields or [])
                    )
                if all(
                    all(bool(v) is False for v in _schema[opt][x].values())
                    for x in range(len(_schema[opt]))
                ):
                    _schema.pop(opt)
                else:
                    _schema[opt] = list(unique_everseen(_schema[opt]))
    return _schema


def relative_path(source_file, target_file):
    return Path(source_file).parent / target_file


def read_definition(file: Path) -> dict[str, Any]:
    "Reads definition from file into a dictionary"
    if isinstance(file, str):
        file = Path(file)
    if file.suffix == ".json":
        with file.open() as fp:
            return json.load(fp)
    elif file.suffix == ".toml":
        with file.open("rb") as fp:
            return tomli.load(fp)
    else:
        raise ValueError(f"Unsupported file format: {file}")


class Parser:
    """Main parser class that loads a specification

    Typical use of this within Python code::

        import adtl

        parser = adtl.Parser(specification)
        print(parser.tables) # list of tables created

        for row in parser.parse().read_table(table):
            print(row)
    """

    def __init__(
        self,
        spec: Union[str, Path, StrDict],
        include_defs: list[str] = [],
        quiet: bool = False,
        verbose: bool = False,
        parallel: bool = False,
    ):
        """Loads specification from spec in format (default json)

        Args:
            spec: Either the specification file to read (as Path or str), or
                the specification loaded into a dictionary
            include_defs: Definition files to include. These are spliced
                directly into the adtl.defs section of the :ref:`specification`.
            quiet: Boolean that switches on the verbosity of the parser, default False
            verbose: Boolean that switches on extra verbosity to show e.g. overwrite warnings, default False
            parallel: Boolean that switches on parallel processing for parsing, default False
        """

        self.data: StrDict = {}
        self.defs: StrDict = {}
        self.fieldnames: dict[str, list[str]] = {}
        self.specfile = None
        self.include_defs = include_defs
        self.validators: StrDict = {}
        self.schemas: StrDict = {}
        self.quiet = quiet
        self.verbose = verbose
        self.parallel = parallel
        self.date_fields = []
        self.report = {
            "validation_errors": defaultdict(Counter),
            "total_valid": defaultdict(int),
            "total": defaultdict(int),
        }
        self.report_available = False
        if isinstance(spec, str):
            spec = Path(spec)
        if isinstance(spec, Path):
            self.specfile = spec
            fmt = spec.suffix[1:]
            if fmt not in SUPPORTED_FORMATS:
                raise ValueError(f"adtl specification format not supported: {fmt}")
            with spec.open("rb") as fp:
                self.spec = SUPPORTED_FORMATS[fmt](fp)
        else:
            self.spec = spec
        self.header = self.spec.get("adtl", {})
        if not self.header:
            raise ValueError("Specification missing required 'adtl' header")
        if self.specfile:
            self.include_defs = [
                relative_path(self.specfile, definition_file)
                for definition_file in self.header.get("include-def", [])
            ] + self.include_defs
        self.defs = self.header.get("defs", {})
        if self.include_defs:
            for definition_file in self.include_defs:
                self.defs.update(read_definition(definition_file))
        self.spec = expand_refs(self.spec, self.defs)

        self.validate_spec()
        for table in (t for t in self.tables if self.tables[t]["kind"] == "oneToMany"):
            self.spec[table] = expand_for(self.spec[table])
        for table in self.tables:
            if self.tables[table].get("groupBy"):
                self.data[table] = defaultdict(dict)
            else:
                self.data[table] = []
            if schema := self.tables[table].get("schema"):
                optional_fields = self.tables[table].get("optional-fields")
                if schema.startswith("http"):
                    try:
                        res = requests.get(schema)
                        if res.status_code != 200:
                            logger.warning(
                                f"Could not fetch schema for table {table!r}, will not "
                                "validate"
                            )
                            continue
                    except ConnectionError:  # pragma: no cover
                        logger.warning(
                            f"Could not fetch schema for table {table!r}, will not "
                            "validate"
                        )
                        continue
                    self.schemas[table] = make_fields_optional(
                        res.json(), optional_fields
                    )
                else:  # local file
                    with (self.specfile.parent / schema).open() as fp:
                        self.schemas[table] = make_fields_optional(
                            json.load(fp), optional_fields
                        )
                self.date_fields.extend(get_date_fields(self.schemas[table]))

        self._set_field_names()
        self.empty_fields = self.header.get("emptyFields", None)

    @lru_cache
    def get_namespace_uuid(self):
        namespace_str = json.dumps(self.header, sort_keys=True)
        toml_hash = hashlib.sha1(namespace_str.encode("utf-8")).hexdigest()
        return uuid.uuid5(uuid.NAMESPACE_DNS, toml_hash)

    @lru_cache
    def ctx(self, attribute: str):
        return {
            "is_date": attribute in self.date_fields,
            "defaultDateFormat": self.header.get(
                "defaultDateFormat", DEFAULT_DATE_FORMAT
            ),
            "skip_pattern": (
                re.compile(self.header.get("skipFieldPattern"))
                if self.header.get("skipFieldPattern")
                else False
            ),
            "returnUnmatched": self.header.get("returnUnmatched", False),
            "namespace": self.get_namespace_uuid(),
        }

    def validate_spec(self):
        "Raises exceptions if specification is invalid"
        for required in ["tables", "name", "description"]:
            if required not in self.header:
                raise ValueError(f"Specification header requires key: {required}")
        self.tables = self.header["tables"]
        self.name = self.header["name"]
        self.description = self.header["description"]

        for table in self.tables:
            aggregation = self.tables[table].get("aggregation")
            group_field = self.tables[table].get("groupBy")
            kind = self.tables[table].get("kind")
            discriminator = self.tables[table].get("discriminator")
            if kind is None:
                raise ValueError(
                    f"Required 'kind' attribute within 'tables' not present for {table}"
                )
            if group_field is not None and aggregation not in [
                "lastNotNull",
                "applyCombinedType",
            ]:
                raise ValueError(
                    f"groupBy needs 'aggregation' to be set for table: {table}"
                )
            if discriminator is None and kind == "oneToMany":
                raise ValueError("discriminator is required for 'oneToMany' tables")

    def validate_row(self, table, row, expanded):
        if (self.tables[table]["kind"] == "oneToMany") and expanded:
            # Need to validate each row against an individual subschema
            attr = row.get(self.tables[table]["discriminator"])
            validator = self.validators[table].get(attr)
            if not validator:
                raise fastjsonschema.JsonSchemaValueException(
                    f"No validator found for attribute '{attr}' in table '{table}'"
                )
            validator(row)
        else:
            self.validators[table](row)

    def _set_field_names(self):
        for table in self.tables:
            if table not in self.spec:
                raise ValueError(
                    f"Parser specification missing required '{table}' element"
                )
            if self.tables[table].get("kind") != "oneToMany":
                self.fieldnames[table] = sorted(list(self.spec[table].keys()))
            else:
                if table not in self.schemas:
                    print(
                        f"Warning: no schema found for {table!r}, field names may be "
                        "incomplete!"
                    )
                    self.fieldnames[table] = list(
                        self.tables[table].get("common", {}).keys()
                    ) + sorted(
                        list(set(sum([list(m.keys()) for m in self.spec[table]], [])))
                    )
                else:
                    self.fieldnames[table] = sorted(self.schemas[table]["properties"])
                if commonMappings := self.tables[table].get("common", {}):
                    for match in self.spec[table]:
                        match.update(commonMappings)

    def _default_if(self, table: str, rule: StrDict):
        """
        Default behaviour for oneToMany table, row not displayed if there's an empty
        string or values not mapped in the rule.
        """

        data_options = [
            option["required"][0] for option in self.schemas[table]["oneOf"]
        ]

        option = set(data_options).intersection(rule.keys()).pop()

        if "combinedType" not in rule[option]:
            field = rule[option]["field"]

            flags = {
                flag: True
                for flag in ("can_skip", "caseInsensitive")
                if flag in rule[option]
            }

            if "values" in rule[option] and not rule[option].get(
                "ignoreMissingKey", False
            ):
                values = rule[option]["values"]
                if_rule = {"any": [{field: v, **flags} for v in values]}
            else:
                if_rule = {field: {"!=": ""}, **flags}
        else:
            assert rule[option]["combinedType"] in [
                "any",
                "all",
                "firstNonNull",
                "set",
                "list",
                "min",
                "max",
            ], f"Invalid combinedType: {rule[option]['combinedType']}"
            rules = rule[option]["fields"]

            def create_if_rule(rule):
                field = rule["field"]
                values = rule.get("values", [])
                flags = {
                    flag: True
                    for flag in ("can_skip", "caseInsensitive")
                    if flag in rule
                }

                if values and not rule.get("ignoreMissingKey", False):
                    if_condition = [{field: v, **flags} for v in values]
                else:
                    if_condition = [{field: {"!=": ""}, **flags}]

                return if_condition

            if_rule = {"any": sum(map(create_if_rule, rules), [])}

        rule["if"] = if_rule
        return rule

    def _parse_row_for_table(self, table: str, row: StrDict):
        """
        Parses a row of data for a given table into the new format

        Args:
            table: Table to update
            row: Dictionary with keys as field names and values as field values
        """

        kind = self.tables[table].get("kind")
        if self.schemas.get(table):
            schema_p = self.schemas[table]["properties"]
        else:
            schema_p = None

        if kind == "oneToMany":
            data = []
            for match in self.spec[table]:
                if "if" not in match:
                    match = self._default_if(table, match)
                if parse_if(row, match["if"], self.ctx):
                    data.append(
                        remove_null_keys(
                            {
                                attr: get_value(
                                    row,
                                    match[attr],
                                    self.ctx(attr),
                                    schema_p[attr].get("type") if schema_p else None,
                                )
                                for attr in set(match.keys()) - {"if"}
                            }
                        )
                    )
            return data
        elif kind == "constant":  # only one row
            return self.spec[table]
        else:  # groupBy
            parsed_row = {}
            for attr in self.spec[table]:
                value = get_value(
                    row,
                    self.spec[table][attr],
                    self.ctx(attr),
                    schema_p[attr].get("type") if schema_p else None,
                )
                if value is not None and value != []:
                    parsed_row[attr] = value
            return remove_null_keys(parsed_row)

    def group_rows(
        self, table: str, group_field: str, aggregation: str, rows: Iterable[StrDict]
    ):
        """
        Applys the 'groupBy' rule and any 'combinedType' rules to the rows of data
        grouped by the group_field (e.g. an ID number).
        """

        def group_attrs(rows, attrs):
            """
            For a grouped set of rows, apply the combinedType rules to the attributes
            """
            if len(rows) == 1:
                return rows[0]  # Return a single dictionary

            combined_row = {}

            for attr in attrs:
                if ("combinedType" in self.spec[table][attr]) and (
                    aggregation == "applyCombinedType"
                ):
                    combined_type = self.spec[table][attr]["combinedType"]
                    values = [
                        row.get(attr) for row in rows if row.get(attr) not in (None, "")
                    ]

                    if not values:
                        continue
                    elif combined_type in ["all", "any", "min", "max"]:
                        combined_row[attr] = eval(combined_type)(values)
                    elif combined_type == "set":
                        combined_row[attr] = list(
                            set(item for sublist in values for item in sublist)
                        )
                    elif combined_type == "list":
                        combined_row[attr] = [
                            item for sublist in values for item in sublist
                        ]
                    elif combined_type == "firstNonNull":
                        combined_row[attr] = values[0]  # First non-null value
                    else:
                        warnings.warn(
                            f"Invalid combinedType: {combined_type} for {attr}",
                            UserWarning,
                        )
                else:
                    data = [
                        row.get(attr)
                        for row in rows
                        if row.get(attr) not in (None, "", [], {})
                    ]
                    if data:
                        if len(data) > 1 and not all(x == data[0] for x in data):
                            if self.verbose:
                                warnings.warn(
                                    f"Multiple rows of data found for {attr} without a"
                                    f" combinedType listed. Data being overwritten: {data}",
                                    UserWarning,
                                )
                            else:
                                logger.debug(
                                    f"Multiple rows of data found for {attr} without a"
                                    " combinedType listed. Data being overwritten."
                                )
                        combined_row[attr] = data[-1]

            return combined_row

        grouped_rows = defaultdict(list)

        # Group rows by the specified field
        for row in rows:
            grouped_rows[row[group_field]].append(row)

        fields = list(self.spec[table].keys())
        fields.remove(group_field)

        # Apply grouping function
        grouped_results = {
            key: remove_null_keys(group_attrs(group, fields))
            for key, group in grouped_rows.items()
        }

        # Convert back to list of dictionaries
        self.data[table] = [
            {group_field: key, **values} for key, values in grouped_results.items()
        ]

    def parse(self, file: str, encoding: str = "utf-8-sig", skip_validation=False):
        """Transform file according to specification

        Args:
            file: Source file to transform
            encoding: Source file encoding
            skip_validation: Whether to skip validation, default off

        Returns:
            adtl.Parser: Returns an instance of itself, updated with the parsed tables
        """
        with open(file, newline="") as f:
            row_count = sum(1 for _ in f) - 1  # Exclude header

        with open(file, encoding=encoding) as fp:
            reader = csv.DictReader(fp)

            def clean_empty_vals(reader, na_values=self.empty_fields):
                for row in reader:
                    yield {k: ("" if v == na_values else v) for k, v in row.items()}

            return self.parse_rows(
                clean_empty_vals(reader) if self.empty_fields else reader,
                Path(file).name,
                row_count,
                skip_validation=skip_validation,
            )

    def parse_rows(
        self,
        rows: Iterable[StrDict],
        file_name: str,
        row_count: float | None = None,
        skip_validation=False,
    ):
        """Transform rows from an iterable according to specification

        Args:
            rows: Iterable of rows, specified as a dictionary of
                    (field name, field value) pairs
            skip_validation: Whether to skip validation, default off

        Returns:
            adtl.Parser: Returns an instance of itself, updated with the parsed tables
        """

        def process_row(row):
            """Process a single row in the data file"""

            row_store = dict.fromkeys(self.tables, None)

            for table in self.tables:
                try:
                    row_store[table] = self._parse_row_for_table(table, row)
                except ValueError:  # pragma: no cover
                    print(
                        "\n".join(
                            [
                                f"{key} = {value}"
                                for key, value in row.items()
                                if value not in ["", None]
                            ]
                        )
                    )
                    raise

            return row_store

        data = Parallel(n_jobs=-1 if self.parallel else 1)(
            delayed(process_row)(row)
            for row in (
                tqdm(
                    rows,
                    desc=f"[{self.name}] parsing {file_name}",
                    total=row_count,
                    disable=self.quiet,
                )
            )
        )

        # merge each row for each table into one data dict per table
        self.data = {
            key: list(values)
            for key, values in zip(data[0], zip(*[d.values() for d in data]))
        }
        for table in self.tables:
            group_field = self.tables[table].get("groupBy")
            aggregation = self.tables[table].get("aggregation")
            if group_field:
                self.group_rows(table, group_field, aggregation, self.data[table])
            if self.tables[table].get("kind") == "oneToMany":
                self.data[table] = list(chain(*self.data[table]))

        self.report_available = not skip_validation
        if not skip_validation:
            for table in self.schemas:
                if self.tables[table]["kind"] == "oneToMany":
                    self.validators[table], attr_schemas = util.expand_schema(
                        self.schemas[table], self.tables[table].get("discriminator")
                    )
                else:
                    self.validators[table] = fastjsonschema.compile(self.schemas[table])
                    attr_schemas = False
                for row in tqdm(
                    self.read_table(table),
                    desc=f"[{self.name}] validating {table} table",
                    disable=self.quiet,
                ):
                    self.report["total"][table] += 1
                    try:
                        self.validate_row(table, row, attr_schemas)
                        row["adtl_valid"] = True
                        self.report["total_valid"][table] += 1
                    except fastjsonschema.exceptions.JsonSchemaValueException as e:
                        row["adtl_valid"] = False
                        row["adtl_error"] = e.message
                        self.report["validation_errors"][table].update([e.message])
        return self

    def clear(self):
        "Clears parser state"
        self.data = {}

    def read_table(self, table: str) -> Iterable[StrDict]:
        """Returns parsed table

        Args:
            table: Table to read

        Returns:
            Iterable of transformed rows in table
        """
        if table not in self.tables:
            raise ValueError(f"Invalid table name: {table}")
        for row in self.data[table]:
            yield row

    def write_csv(self, table: str, output: str | None = None) -> str | None:
        """Writes to output as CSV a particular table

        Args:
            table: Table that should be written to CSV
            output: (optional) Output file name. If not specified, defaults to parser
                    name + table name with a csv suffix.
        """

        def writerows(fp, table):
            writer = csv.DictWriter(
                fp,
                fieldnames=(
                    ["adtl_valid", "adtl_error"] if table in self.validators else []
                )
                + self.fieldnames[table],
            )
            writer.writeheader()
            for row in self.read_table(table):
                writer.writerow(row)
            return fp

        if output:
            with open(output, "w") as fp:
                writerows(fp, table)
            return None
        else:
            buf = io.StringIO()
            return writerows(buf, table).getvalue()

    def write_parquet(self, table: str, output: str | None = None) -> str | None:
        """Writes to output as parquet a particular table

        Args:
            table: Table that should be written to parquet
            output: (optional) Output file name. If not specified, defaults to parser
                    name + table name with a parquet suffix.
        """

        # Read the table data
        data = list(self.read_table(table))

        # Convert data to Pandas DataFrame
        df = pd.DataFrame(data)

        if table in self.validators:
            valid_cols = [c for c in ["adtl_valid", "adtl_error"] if c in df.columns]
            df_validated = df[
                valid_cols
                + [
                    *[
                        col
                        for col in df.columns
                        if (col != "adtl_valid" and col != "adtl_error")
                    ],  # All other columns, in their original order
                ]
            ]
        else:
            df_validated = df

        if output:
            df_validated.to_parquet(output)
        else:
            buf = io.BytesIO()
            df_validated.to_parquet(buf)
            return buf.getvalue()

    def show_report(self):
        "Shows report with validation errors"
        if self.report_available:
            print("\n|table       \t|valid\t|total\t|percentage_valid|")
            print("|---------------|-------|-------|----------------|")
            for table in self.report["total"]:
                print(
                    f"|{table:14s}\t|{self.report['total_valid'][table]}\t"
                    f"|{self.report['total'][table]}\t"
                    f"|{self.report['total_valid'][table] / self.report['total'][table]:%} |"  # noqa:E501
                )
            print()
            for table in self.report["validation_errors"]:
                print(f"## {table}\n")
                for message, count in self.report["validation_errors"][
                    table
                ].most_common():
                    print(f"* {count}: {message}")
                print()

    def save(
        self, output: str | None = None, format: Literal["csv", "parquet"] = "csv"
    ):
        """Saves all tables to CSV

        Args:
            output: (optional) Filename prefix that is used for all tables
        """

        if format == "parquet":
            for table in self.tables:
                self.write_parquet(table, f"{output}-{table}.parquet")

        elif format == "csv":
            for table in self.tables:
                self.write_csv(table, f"{output}-{table}.csv")

        else:
            raise ValueError(f"'Parser.save()': Invalid format: {format}")



"""
Mixin class for handling long table mappings.
"""

from __future__ import annotations

from functools import cached_property
from typing import Protocol

import pandas as pd

from .config.config import Config


class LongTableMixin(Protocol):
    """
    Mixin providing access to common long-table config fields.

    Requires:
        - self.config: dict-like config with a "long_tables" section
        - self.name: str, the name of the current table
        - self.schema_fields: The schema 'properties' list for the current table
    """

    config: Config
    name: str
    schema_fields: list[dict]

    @cached_property
    def common_cols(self) -> list[str]:
        """Returns the common columns for the long table"""
        ccs = self.config.long_tables[self.name].common_cols
        if not ccs:
            ccs = list(self.config.long_tables[self.name].common_fields.keys())

        return ccs

    @property
    def common_fields(self) -> pd.Series:
        if not hasattr(self, "_common_fields"):
            self._common_fields = self.config.long_tables[self.name].common_fields

        return self._common_fields

    @cached_property
    def variable_col(self) -> str:
        """Returns the variable column for the long table"""
        return self.config.long_tables[self.name].variable_col

    @cached_property
    def value_cols(self) -> list[str]:
        """Returns the value columns for the long table"""
        return self.config.long_tables[self.name].value_cols

    @cached_property
    def other_fields(self) -> list[str]:
        """Returns the other fields in the schema that are not target fields"""
        return [
            f
            for f in self.schema_fields
            if f
            not in [
                *self.common_fields.keys(),
                self.variable_col,
                *self.value_cols,
            ]
        ]



from __future__ import annotations

import abc
from functools import cached_property
from typing import Union

import numpy as np
import pandas as pd

from ..config.config import get_config
from ..dict_reader import format_dict
from ..util import read_json


class BaseMapper(abc.ABC):
    """
    Abstract class for creating an intermediate mapping file linking the data dictionary to
    a schema's fields and values. Will be used to create Wide and Long mapping formats.

    Use `create_mapping()` to write out the mapping file, as the function equivalent
    of the command line `create-mapping` script.

    Parameters
    ----------
    data_dictionary
        The data dictionary to use
    table_name
        The name of the table to map to
    """

    INDEX_FIELD = "target_field"

    def __init__(self, data_dictionary: Union[str, pd.DataFrame], table_name: str):
        self.name = table_name

        self.config = get_config()

        self.language = self.config.language
        self.model = self.config._llm

        if self.model is None:
            self.config.check_llm_setup()

        self.schema = read_json(self.config.schemas[table_name])
        self.schema_fields = self.schema["properties"]

        self.data_dictionary = format_dict(data_dictionary)

    # Abstract methods ---------------------------------------

    @abc.abstractmethod
    def _iter_value_tuples(self) -> list[tuple[str, list[str], list[str]]]: ...

    @abc.abstractmethod
    def create_mapping(self, save=True, file_name="mapping_file") -> pd.DataFrame: ...

    # Common properties ---------------------------------------

    @cached_property
    def common_values(self) -> pd.Series:
        """
        Returns the commonly repeated values in the source data
        Usually this indicates that the source field is an enum or boolean
        """
        if "common_values" in self.data_dictionary.columns:
            cv = self.data_dictionary.common_values
            cv.index = self.data_dictionary.source_field
            return cv
        elif "choices" in self.data_dictionary.columns:
            choices = self.data_dictionary.choices
            cv = choices.apply(
                lambda x: list(x.values()) if isinstance(x, dict) else None
            )
            cv.index = self.data_dictionary.source_field
            return cv
        else:
            # pandera schema validation means this should never be reached
            raise ValueError(
                "No common values or choices column found in data dictionary"
            )  # pragma: no cover

    @cached_property
    def common_values_mapped(self) -> pd.Series:
        try:
            filtered_dict = self.filtered_data_dict
        except AttributeError:
            raise AttributeError(
                "fields have to be mapped using the `match_fields_to_schema` method"
                " first"
            )
        cv = self.common_values
        return filtered_dict.source_field.apply(
            lambda x: cv.loc[x] if isinstance(x, str) else None
        )

    @property
    def mapped_fields(self) -> pd.Series:
        try:
            return self._mapped_fields
        except AttributeError:
            raise AttributeError(
                "mapped_fields have to be created using `match_fields_to_schema` method"
            )

    @mapped_fields.setter
    def mapped_fields(self, value: pd.Series):
        self._mapped_fields = value

    def _relabel_choices(self, map_df) -> pd.DataFrame:
        """
        If 'choices' are present in the data dictionary, relabel the choices
        in the mapping dataframe to match the source field names.
        e.g.
        if the data dictionary indicates {1: Man, 2:Female, 3:Unknown}, and the LLM maps
        {Man:male, Female:female, Unknown:unknown}, then `relabel_choices` should give
        {1:male, 2:female, 3:unknown}
        """

        filtered_dict = self.data_dictionary[
            self.data_dictionary["source_field"].isin(map_df["source_field"])
        ][["source_field", "choices"]]

        mapping_choices = map_df[["source_field", "value_mapping"]]

        choices = filtered_dict.merge(mapping_choices, on="source_field")
        choices["combined_choices"] = choices.apply(
            lambda x: (
                {k: x["value_mapping"].get(v) for k, v in x["choices"].items()}
                if isinstance(x["choices"], dict)
                else None
            ),
            axis=1,
        )

        merged_mapping_dict = (
            map_df.reset_index()
            .merge(
                choices[["source_field", "combined_choices"]],
                on="source_field",
                how="left",
            )
            .drop_duplicates(subset="target_field")
            .set_index("target_field")
        )
        mmd = merged_mapping_dict.drop(columns=["value_mapping"]).rename(
            columns={"combined_choices": "value_mapping"}
        )

        return mmd

    def post_process_mapping(self, mapping_dict, save, file_name) -> pd.DataFrame:
        """
        Turn lists & dicts into strings for consistency with saved CSV, then save.
        """

        if "choices" in self.data_dictionary:
            mapping_dict = self._relabel_choices(mapping_dict)

        for col in ["common_values", "target_values"]:
            if col in mapping_dict.columns:
                mapping_dict[col] = mapping_dict[col].apply(
                    lambda x: (
                        " | ".join(str(item) for item in x)
                        if isinstance(x, (list, np.ndarray))
                        else x
                    )
                )
        for col in ["choices", "value_mapping"]:
            if col in mapping_dict.columns:
                mapping_dict[col] = mapping_dict[col].apply(
                    lambda x: (
                        " | ".join(f"{k}={v}" for k, v in x.items())
                        if isinstance(x, dict)
                        else x
                    )
                )

        if save is False:
            return mapping_dict
        else:
            # Write to CSV
            if not file_name.endswith(".csv"):
                file_name += ".csv"
            mapping_dict.to_csv(file_name)
            return mapping_dict

    def match_values_to_schema(self) -> pd.DataFrame:
        """
        Use the LLM to match the common values from the data dictionary to the target
        values in the schema - i.e. enum or boolean options.
        """

        values_tuples = list(self._iter_value_tuples())

        # to LLM
        value_pairs = self.model.map_values(values_tuples, self.language)

        value_mapping = {}

        for p in value_pairs.values:
            f = p.field_name
            value_dict = {
                pair.source_value: pair.target_value for pair in p.mapped_values
            }
            value_mapping[f] = value_dict

        self.mapped_values = pd.Series(value_mapping, name="value_mapping")
        self.mapped_values.index.name = self.INDEX_FIELD

        return self.mapped_values



from __future__ import annotations

import warnings
from enum import Enum
from functools import cached_property
from typing import Optional

import numpy as np
import pandas as pd
from pydantic import BaseModel, create_model

from ..mixin import LongTableMixin
from .base_mapper import BaseMapper


class LongMapper(BaseMapper, LongTableMixin):
    """
    Class for creating an intermediate mapping file linking the data dictionary to
    long-format schema's fields and values.
    """

    INDEX_FIELD = "source_field"

    @cached_property
    def common_values_mapped(self) -> pd.Series:
        try:
            filtered_dict = self.filtered_data_dict
        except AttributeError:
            raise AttributeError(
                "Fields have to be mapped using the `match_fields_to_schema` method"
                " first"
            )
        cv = self.common_values
        return cv.loc[filtered_dict[filtered_dict["variable_name"].notna()].index]

    @cached_property
    def target_values(self) -> pd.Series:
        """Returns the enum values or boolean options for the target schema"""

        def _value_options(f):
            if "boolean" in self.schema_fields[f].get("type", ["string", "null"]):
                return ["True", "False", "None"]
            elif "string" in self.schema_fields[f].get("type", ["string", "null"]):
                return self.schema_fields[f].get("enum", np.nan)
            elif "array" in self.schema_fields[f].get("type", ["string", "null"]):
                return self.schema_fields[f].get("items", {}).get("enum", np.nan)
            else:
                return np.nan

        return pd.Series(
            {f: _value_options(f) for f in self.value_cols},
            self.value_cols,
        )

    def _iter_value_tuples(self):
        for f in self.mapped_fields:
            s = self.common_values_mapped.get(f)
            t = self.target_values.get(self.filtered_data_dict.loc[f, "value_col"])
            if s is not None and t is not None:
                yield (f, s, t)

    def _create_data_model(self) -> BaseModel:
        def _enum_creator(name: str, enums: list[str]) -> dict:
            """
            Creates a dictionary for a single entry in the long table format.
            """
            return Enum(name, {v.upper(): v for v in enums})

        fields = {
            "source_description": (str, ...),
            "variable_name": (str, ...),
            "value_col": (
                Optional[_enum_creator("ValueColEnum", self.value_cols)],
                None,
            ),
        }

        if self.schema_fields[self.variable_col].get("enum", []):
            VarColEnum = _enum_creator(
                "VarColEnum", self.schema_fields[self.variable_col]["enum"]
            )
            fields["variable_name"] = (Optional[VarColEnum], None)

        # Add arbitrary fields from CSV headers
        for field in self.other_fields:
            if self.schema_fields[field].get("enum", []):
                # If the field has an enum, create an Enum type for it
                EnumType = _enum_creator(
                    f"{field}Enum", self.schema_fields[field]["enum"]
                )
                fields[field] = (Optional[EnumType], None)
            else:
                fields[field] = (Optional[str], None)

        SingleEntry = create_model("SingleEntry", **fields)
        return SingleEntry

    def _check_config(self):
        """
        Check that the config file has the correct fields for the long table.
        """
        if self.config.long_tables is None:
            raise ValueError(
                "No long tables defined in config file. Please set 'long_tables' in the config file."
            )

        if self.schema_fields[self.variable_col].get("enum") is None:
            raise ValueError(
                f"Variable column '{self.variable_col}' in schema does not have an enum set. "
                "Please set 'enum' in the schema for this field."
            )

    def set_common_fields(self, common_fields: dict[str, str]):
        """
        Function to assign fields to the common fields of the long table - i.e. fields
        which should be filled by the same text or source field in every row of the long
        table.
        """
        if self.common_cols != list(common_fields.keys()):
            raise ValueError(
                f"Common columns {self.common_cols} set in the config file do not"
                f" match provided common fields {common_fields.keys()}"
            )

        self._common_fields = common_fields
        self._common_cols = list(common_fields.keys())

    def match_fields_to_schema(self) -> pd.DataFrame:
        """
        Use the LLM to match the target (schema) fields to the descriptions of the
        source data fields from the data dictionary.
        """

        data_format = self._create_data_model()

        if self.common_cols:
            self.uncommon_data_dict = self.data_dictionary[
                ~self.data_dictionary.source_field.isin(self.common_cols)
            ].drop(columns=["source_type"])
        else:
            self.uncommon_data_dict = self.data_dictionary.drop(columns=["source_type"])

        source_descriptions = self.uncommon_data_dict.source_description

        mappings = self.model.map_long_table(
            data_format,
            source_descriptions.tolist(),
            self.schema_fields[self.variable_col]["enum"],
        )

        mapping_dict = pd.DataFrame(mappings.model_dump(mode="json")["long_table"])

        missed_descrips = len(source_descriptions) != len(mapping_dict)

        if missed_descrips:
            mapping_dict = pd.merge(
                pd.DataFrame({"source_description": source_descriptions}),
                mapping_dict,
                how="left",
                on="source_description",
            )

            assert len(source_descriptions) == len(mapping_dict), (
                "malformed descriptions!"
            )

        df_merged = pd.merge(
            self.uncommon_data_dict,
            mapping_dict,
            how="left",
            on="source_description",
        )

        df_merged.set_index("source_field", inplace=True, drop=True)

        self.mapped_fields = df_merged.index
        self.filtered_data_dict = df_merged
        return df_merged

    def create_mapping(
        self,
        save=True,
        file_name="mapping_file",
    ) -> pd.DataFrame:
        """
        Creates an intermediate mapping dataframe linking the data dictionary to schema
        fields. The index contains the source field names, and the columns are:
        * source_description
        * common_values OR choices (depending on the data dictionary)
        * <variable_name> (the name of the column identified in the config file)
        * value_col
        * any other fields in the long schema.

        Raises a warning if any fields are present in the schema where a
        corresponding source field in the data dictionary has not been found.

        Parameters
        ----------
        save
            Whether to save the mapping to a CSV file. If True, lists and dicts are
            converted to strings before saving.
        file_name
            The name to use for the CSV file
        """

        self._check_config()

        if self.common_cols and not self.common_fields:
            raise ValueError(
                "Common fields must be set in the config file or set using the"
                " `set_common_fields` method before mapping."
            )

        mapping_dict = self.match_fields_to_schema()

        unmapped = mapping_dict[mapping_dict["variable_name"].isna()].index
        if any(unmapped):
            warnings.warn(
                f"The following fields have not been mapped to the new schema: {list(unmapped)}",
                UserWarning,
            )

        mapped_vals = self.match_values_to_schema()

        mapping_dict = pd.concat([mapping_dict, mapped_vals], axis=1)

        # Add in the common columns to the file
        for col, value in self.common_fields.items():
            mapping_dict[col] = value

        mapping_dict.rename(columns={"variable_name": self.variable_col}, inplace=True)

        return self.post_process_mapping(mapping_dict, save=save, file_name=file_name)



"""
Generate TOML parser from intermediate CSV file
"""

from __future__ import annotations

import abc
import argparse
import json
import logging
from functools import cached_property
from pathlib import Path
from typing import Any, Union

import pandas as pd

from .config.config import get_config, setup_config
from .mixin import LongTableMixin
from .toml_writer import dump
from .util import DEFAULT_CONFIG, parse_llm_mapped_values, read_data, read_json

logger = logging.getLogger(__name__)
INPUT_FORMAT = Union[pd.DataFrame, str, Path]


class TableParser(abc.ABC):
    INDEX_FIELD = "target_field"

    def __init__(self, mapping: pd.DataFrame, schema: dict, table_name: str):
        self.mapping = mapping
        self.schema = schema
        self.name = table_name
        self.config = get_config()

    def single_field_mapping(self): ...

    def make_toml_table(self) -> dict[str, Any]: ...

    @cached_property
    def constant_field(self) -> dict[str, bool]: ...

    @property
    def schema_fields(self):
        """Returns all the fields for `table` and their properties"""
        return self.schema["properties"]

    @cached_property
    def field_types(self) -> dict[str, list[str]]:
        """Returns the field types of the target schema"""
        s = self.schema_fields
        return {f: s[f].get("type", ["string", "null"]) for f in s}

    @cached_property
    def parsed_choices(self) -> pd.Series:
        """Returns the mapped values for each target field"""
        values = self.mapping.value_mapping.map(parse_llm_mapped_values)
        values.index = self.mapping[self.INDEX_FIELD]
        return values

    def update_constant_fields(self, fields: dict[str, bool]) -> None:
        """Update the constant fields"""
        for field, value in fields.items():
            if field not in self.constant_field:
                raise ValueError(f"Field '{field}' is not a valid schema field.")
            if value not in [True, False]:
                raise ValueError(f"Value for field '{field}' must be True or False.")
            self.constant_field[field] = value


class WideTableParser(TableParser):
    """
    Class for generating a wide table from the mappings.
    """

    @cached_property
    def constant_field(self) -> dict[str, bool]:
        """
        If a column in the mapping file should be pulled from a schema field, True,
        otherwise False
        """
        return {col: False for col in self.schema_fields}

    @cached_property
    def references_definitions(self) -> tuple[dict[str, str], dict[str, dict]]:
        """Finds and returns the references and definitions for the mappings"""
        # use value_counts() on parsed_choices normalise various flavours of Y/N/NK
        value_counts = self.parsed_choices.value_counts()
        return self.refs_defs(value_counts, self.config.num_refs)

    def refs_defs(self, choices, num_refs):
        references = {}
        definitions = {}

        top_mappings = choices[choices > 1][:num_refs].index

        # only add one boolean map for simplicity
        boolean_map_found = False
        for mapping in top_mappings:
            if boolean_map_found and True in mapping.values():
                continue
            if True in mapping.values():
                references[json.dumps(mapping, sort_keys=True)] = "Y/N/NK"
                definitions["Y/N/NK"] = {
                    "caseInsensitive": True,
                    "values": mapping,
                }
                boolean_map_found = True
                continue
            c = mapping
            name = "/".join(map(str, c.values()))
            references[json.dumps(mapping, sort_keys=True)] = name
            definitions[name] = {"values": c, "caseInsensitive": True}

        return references, definitions

    def single_field_mapping(self, match: pd.DataFrame) -> dict[str, Any]:
        """Make a single field mapping from a single row of the mappings dataframe"""

        choices = self.parsed_choices[match.target_field]

        out = {"field": match.source_field, "description": match.source_description}
        references = self.references_definitions[0]
        if choices:
            if (choice_key := json.dumps(choices, sort_keys=True)) in references:
                out["ref"] = references[choice_key]
            else:
                out["values"] = choices
                out["caseInsensitive"] = True

            if "array" in self.field_types[match.target_field]:
                out["type"] = "enum_list"

        return out

    def make_toml_table(self) -> dict[str, Any]:
        """Make single TOML table from mappings"""

        outmap = {}

        for field, field_matches in self.mapping.groupby("target_field"):
            if len(field_matches) == 1:  # single field
                if not any(field_matches["source_field"].isna()):
                    outmap[field] = self.single_field_mapping(field_matches.iloc[0])

            else:  # pragma: no cover
                raise NotImplementedError("CombinedType not supported")

        # check for missing required fields
        if "required" in self.schema:
            for field in self.schema["required"]:
                if field not in outmap:
                    logger.warning(
                        f"Missing required field {field} in {self.name} schema."
                        " Adding empty field..."
                    )
                    outmap[field] = ""

        return {self.name: outmap}, self.references_definitions[1]


class LongTableParser(TableParser, LongTableMixin):
    """
    Class for generating a long table from the mappings.
    """

    INDEX_FIELD = "source_field"

    @cached_property
    def constant_field(self) -> dict[str, bool]:
        """If a column in the mapping file should be pulled from a schema field, True, otherwise False"""
        config = {col: False for col in self.schema_fields}
        config[self.variable_col] = True

        for col in self.other_fields:
            config[col] = True

        return config

    def _validate_mapping(self):
        """Validate the mapping dataframe for the long table"""
        if any(self.mapping[self.variable_col].isna()):
            raise ValueError(
                f"Mapping dataframe must not contain NaN values in '{self.variable_col}' column."
            )
        if any(self.mapping["value_col"].isna()):
            raise ValueError(
                "Mapping dataframe must not contain NaN values in the 'value_col' column."
            )

    def single_entry_mapping(self, data: pd.DataFrame) -> dict[str, Any]:
        """Make a single entry mapping from a single row of the mappings dataframe"""

        def add_field(field, text: str) -> Any:
            """Check if a field should be added to the output"""
            if self.constant_field.get(field, False):
                return text
            return {"field": text}

        out = {
            self.variable_col: data[self.variable_col],
            data.value_col: add_field(data.value_col, data.source_field),
            **{field: add_field(field, data[field]) for field in self.common_cols},
        }

        choices = self.parsed_choices[data.source_field]
        if choices:
            out[data.value_col].update(
                {
                    "values": choices,
                    "caseInsensitive": True,
                }
            )

        for field in self.other_fields:
            if not pd.isna(data[field]):
                out[field] = add_field(field, data[field])

        return out

    def make_toml_table(self) -> dict[str, Any]:
        """Make single TOML table from mappings"""

        self._validate_mapping()

        outmap = []

        for _, row in self.mapping.iterrows():
            outmap.append(self.single_entry_mapping(row))

        return {self.name: outmap}, None


class ParserGenerator:
    """
    Class for creating a TOML parser from an intermediate CSV file.

    Use `create_parser()` to write out the TOML parser file, as the function equivalent
    of the command line `create-parser` script.

    Parameters
    ----------
    mappings : pd.DataFrame | str | Path
        The intermediate CSV file created by `create_mapping.py`
    schema_path : Path | str
        The path to the folder containing all the schema files
    parser_name : str
        The name of the parser
    description : str, optional
        The description of the parser
    constant_fields : dict[str, dict[str, bool]], optional
        Constant fields are those which are single values, rather than taken from a field from the source data.
        For example, if an entire dataset is from the DRC, but a country field is in the target schema, there may not be a
        field in the dataset stating the country.
        A dictionary of constant fields for each table, where the keys are the table names
        and the values are boolean True/False values indicating whether the field should be pulled from the source data
        or not. All fields in wide tables default to False, while long tables default to True for all columns except
        the value column(s).
    """

    def __init__(
        self,
        mappings: Union[INPUT_FORMAT, dict[str, INPUT_FORMAT]],
        schema_path: Union[Path, str],
        parser_name: str,
        description: Union[str, None] = None,
        constant_fields: Union[dict[str, dict[str, bool]], None] = None,
    ):
        if not isinstance(mappings, dict):
            # if not a dict, assume a single mapping file
            mappings = read_data(mappings, "A mapping file")

        self.schema_path = Path(schema_path)
        self.parser_name = parser_name
        self.parser_description = description or parser_name

        self.config = get_config()
        self.tables = list(self.config.schemas.keys())

        if len(self.tables) == 1:
            # if only one table, use the singular form
            self.mappings = {self.tables[0]: mappings}
        else:
            for table in self.tables:
                if table not in mappings:
                    raise ValueError(
                        f"Mapping for table '{table}' not found in provided mappings."
                    )
            self.mappings = {
                table: read_data(mappings[table], "A mapping file")
                for table in self.tables
            }

        self.schemas = {t: read_json(self.config.schemas[t]) for t in self.tables}

        self.table_types = {
            t: "wide" if "target_field" in m.columns else "long"
            for t, m in self.mappings.items()
        }

        self.constant_fields = constant_fields or {}

    def header(self) -> dict[str, Any]:
        "The ADTL-specific header for the TOML file"
        schemas = {}
        for table in self.schemas:
            schemas[table] = {
                "kind": (
                    "oneToOne" if self.table_types[table] == "wide" else "oneToMany"
                ),
                "schema": f"{self.schema_path / Path(self.config.schemas[table])}",
            }

        return {
            "adtl": {
                "name": self.parser_name,
                "description": self.parser_description,
                "returnUnmatched": True,
                "tables": schemas,
                "defs": {},
            }
        }

    def make_single_parser(self) -> dict[str, Any]:
        """
        Takes the csv mapping file from `create_mapping` and writes out a TOML parser

        Generates a TOML parser for use with ADTL using the intermediate CSV file from
        by `create_mapping`. This will generate a TOML file that can be used to parse
        raw data into the format expected by the schema.

        Returns
        -------
        dict
            Dictionary containing the TOML parser data, ready to be written out.
        """

        data = self.header()
        for table in self.tables:
            if self.table_types[table] == "wide":
                parser_class = WideTableParser
            else:
                parser_class = LongTableParser
            parser_class = parser_class(
                self.mappings[table],
                self.schemas[table],
                table,
            )
            if self.constant_fields:
                parser_class.update_constant_fields(self.constant_fields.get(table, {}))

            table_parser, references = parser_class.make_toml_table()
            data.update(table_parser)
            if references:
                data["adtl"]["defs"].update(references)
        return data

    def write_toml(self, data: dict[str, Any], output: str = None):
        """
        Write a dictionary structure to a TOML file, using `output` as the filename if
        provided.

        Parameters
        ----------
        data : dict
            Dictionary containing the TOML parser data
        output : str, optional
            Filename to write the TOML data to. Defaults to the name of the parser.
        """
        if not output:
            output = f"{self.parser_name}.toml"
        with open(output, "wb") as fp:
            dump(data, fp)

    def create_parser(self, file_name: str = None):
        """
        Main function to create the TOML parser from the intermediate CSV file.
        """
        toml_data = self.make_single_parser()
        self.write_toml(toml_data, output=file_name)


def create_parser(
    mappings: Union[pd.DataFrame, str],
    schema_path: Path,
    parser_name: str,
    description: Union[str, None] = None,
    constant_fields: Union[dict[str, dict[str, bool]], None] = None,
):
    """
    Takes the csv mapping file created by `create_mapping` and writes out a TOML parser

    Generates a TOML parser for use with ADTL from the intermediate CSV file generated
    by `create_mapping`. This will generate a TOML file that can be used to parse raw
    data into the format expected by the schema.

    Parameters
    ----------
    mappings
        Path to the CSV file containing the mappings
    schema_path
        Path to the schema file
    parser_name
        Name of the parser to create
    description
        Description of the parser. Defaults to the parser name.
    constant_fields
        Constant fields are those which are single values, rather than taken from a
        field in the source data.

    Returns
    -------
    None
    """
    ParserGenerator(
        mappings,
        schema_path,
        parser_name,
        description,
        constant_fields=constant_fields,
    ).create_parser(parser_name)


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Make TOML from intermediate CSV file created by create_mapping.py",
        prog="autoparser create-parser",
    )
    parser.add_argument("mappings", help="Mapping file to create TOML from", type=str)
    parser.add_argument("schema_path", help="Path where schemas are located")
    parser.add_argument(
        "-o",
        "--output",
        help="Name of the parser to output (default=globalhealth)",
        default="globalhealth",
    )
    parser.add_argument("--description", help="Description of the parser")
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )
    args = parser.parse_args(argv)

    schema_path = Path(args.schema_path)

    setup_config(args.config or DEFAULT_CONFIG)

    ParserGenerator(
        args.mappings,
        schema_path,
        args.output,
        args.description or None,
    ).create_parser()


if __name__ == "__main__":
    main()  # pragma: no cover



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any, Literal, Optional

import tomli
from pydantic import BaseModel, SecretStr, field_validator, model_validator
from typing_extensions import Self

from adtl.autoparser.language_models.base_llm import LLMBase
from adtl.autoparser.language_models.gemini import GeminiLanguageModel
from adtl.autoparser.language_models.openai import OpenAILanguageModel


class ColumnMappingConfig(BaseModel):
    """
    Configuration for column mapping in ADTL autoparser.
    This class defines the structure and fields required for column mapping.
    """

    source_field: str
    source_type: str
    source_description: str
    common_values: Optional[str] = None
    choices: Optional[str] = None

    @model_validator(mode="after")
    def check_common_values_and_choices(self) -> Self:
        if self.common_values is not None and self.choices is not None:
            raise ValueError(
                "Only one from 'common values' and 'choices' can be set at once"
            )

        elif self.common_values is None and self.choices is None:
            raise ValueError(
                "Either 'common values' or 'choices' must be set in column mappings"
            )
        return self


class DefaultColumnMappingConfig(ColumnMappingConfig):
    """
    Default configuration for column mapping in ADTL autoparser.
    This class provides default values for the column mapping fields.
    """

    source_field: str = "Field Name"
    source_type: str = "Field Type"
    source_description: str = "Description"
    common_values: Optional[str] = "Common Values"


class LongTableConfig(BaseModel):
    """
    Configuration for long table mapping in ADTL autoparser.
    This class defines the structure and fields required for long table mapping.
    """

    variable_col: str
    value_cols: list[str]
    common_cols: Optional[list[str]] = []
    common_fields: Optional[dict[str, str]] = {}

    @model_validator(mode="after")
    def check_common_cols_fields(self) -> Self:
        if self.common_cols and self.common_fields:
            raise ValueError(
                "Only one from 'common_cols' and 'common_fields' can be set at once"
            )
        return self


class Config(BaseModel):
    name: str = "autoparser"
    description: str = "Configuration for ADTL autoparser"
    language: str
    schemas: dict[str, str]
    column_mappings: ColumnMappingConfig = DefaultColumnMappingConfig()
    llm_provider: Optional[Literal["openai", "gemini"]] = None
    llm_model: Optional[str] = None
    api_key: Optional[SecretStr] = None
    choice_delimiter: str = ","
    choice_delimiter_map: str = "="
    num_refs: int = 3
    max_common_count: int = 25
    min_common_frequency: Optional[float] = None
    long_tables: Optional[dict[str, LongTableConfig]] = None

    _llm: Optional[LLMBase]

    def model_post_init(self, context: Any) -> None:
        """
        Set up the LLM.
        """
        if self.api_key and (self.llm_provider or self.llm_model):
            kwargs = {"api_key": self.api_key.get_secret_value()}
            if self.llm_model is not None:
                kwargs["model"] = self.llm_model

            if (
                self.llm_provider == "openai"
                or self.llm_model in OpenAILanguageModel.valid_models()
            ):
                self._llm = OpenAILanguageModel(**kwargs)
            elif (
                self.llm_provider == "gemini"
                or self.llm_model in GeminiLanguageModel.valid_models()
            ):
                self._llm = GeminiLanguageModel(**kwargs)
            else:
                raise ValueError(
                    f"Could not set up LLM with provider '{self.llm_provider}' and model '{self.llm_model}'."
                )
        else:
            self._llm = None

    @field_validator("api_key", mode="after")
    @classmethod
    def retrieve_api_key(cls, k) -> Optional[SecretStr]:
        try:
            return SecretStr(os.environ[k.get_secret_value()])
        except KeyError:
            return k

    @model_validator(mode="after")
    def check_common_cols_fields(self) -> Self:
        if self.long_tables:
            for table in self.long_tables.keys():
                if table not in self.schemas:
                    raise ValueError(
                        f"Table '{table}' in 'long_tables' not found in 'schemas'"
                    )
        return self

    def check_llm_setup(self) -> None:
        """
        Check if the LLM is set up correctly.
        Raises an error if the LLM is not configured, and points to the potential error.
        """
        if self._llm is None:
            if self.api_key is None:
                raise ValueError("Config: API key required to set up an LLM")
            if self.llm_provider is None and self.llm_model is None:
                raise ValueError(
                    "Config: LLM provider or model must be specified to set up an LLM"
                )


def _config():
    _config_instance: Optional[Config] = None

    def get_config() -> Config:
        """Returns the current config instance, if set."""
        if _config_instance is None:
            raise RuntimeError("Config not initialized. Call setup_config() first.")
        return _config_instance

    def setup_config(path: Path | str | dict) -> Config:
        """Initializes the config singleton from a file."""

        if isinstance(path, dict):
            data = path
        else:
            path = Path(path)
            if path.suffix == ".json":
                with path.open() as fp:
                    data = json.load(fp)
            elif path.suffix == ".toml":
                with path.open("rb") as fp:
                    data = tomli.load(fp)
            else:
                raise ValueError(f"Unsupported config file format: {path.suffix}")

        nonlocal _config_instance
        _config_instance = Config(**data)

    return get_config, setup_config


get_config, setup_config = _config()



"""
Easy access interface for creating mappings between data dictionaries and schemas.
"""

from __future__ import annotations

import argparse
from pathlib import Path
from typing import Literal, Union

import pandas as pd

from ..config.config import setup_config
from ..util import DEFAULT_CONFIG
from .long_mapper import LongMapper
from .wide_mapper import WideMapper

CONFIG = "../" + DEFAULT_CONFIG


def create_mapping(
    data_dictionary: Union[str, pd.DataFrame],
    table_name: str,
    save: bool = True,
    file_name: str = "mapping_file",
    table_format: Literal["wide", "long"] = "wide",
) -> pd.DataFrame:
    """
    Creates a csv containing the mapping between a data dictionary and a schema.

    Takes a data dictionary and matches both the source fields, and any common values
    to the schema. Uses an LLM to first match the source fields to appropriate schema
    targets, and then to match the common values to appropriate enum or boolean options.

    Parameters
    ----------
    data_dictionary
        Path to a CSV or XLSX file, or a DataFrame, containing the data dictionary.
    table_name
        Name of the table being mapped.
    save
        Whether to save the mapping to a CSV file.
    file_name
        Name of the file to save the mapping to.
    table_format
        Format of the table to create, either 'wide' or 'long'.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the mapping between the data dictionary and the schema.
    """
    if table_format == "long":
        MapperClass = LongMapper
    elif table_format == "wide":
        MapperClass = WideMapper
    else:
        raise ValueError(
            f"Invalid table format: {table_format}. Must be either 'wide' or 'long'."
        )
    df = MapperClass(data_dictionary, table_name).create_mapping(
        save=save, file_name=file_name
    )

    return df


def main(argv=None):
    parser = argparse.ArgumentParser(
        description=(
            "Generate intermediate CSV used by make_toml.py (create-parser) "
            "to create TOML"
        ),
        prog="autoparser create-mapping",
    )
    parser.add_argument("dictionary", help="Data dictionary to use")
    parser.add_argument("table_name", help="Name of the table being mapped")
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={CONFIG})",
        type=Path,
    )
    parser.add_argument(
        "-o", "--output", help="Name to use for output files", default="mapping_file"
    )
    parser.add_argument(
        "--long-table", help="The target table has a long format", action="store_true"
    )
    args = parser.parse_args(argv)

    setup_config(args.config or CONFIG)

    create_mapping(
        data_dictionary=args.dictionary,
        table_name=args.table_name,
        save=True,
        file_name=args.output,
        table_format="long" if args.long_table else "wide",
    )


if __name__ == "__main__":
    main()  # pragma: no cover



from __future__ import annotations

from typing import Any, Optional

import numpy as np
import pandera.pandas as pa
from pandera.typing import Series


class GeneratedDict(pa.DataFrameModel):
    field_name: Series[str] = pa.Field(unique=True, alias="Field Name")
    description: Series[str] = pa.Field(nullable=True, alias="Description")
    field_type: Series[str] = pa.Field(alias="Field Type")
    common_values: Optional[Series[str]] = pa.Field(
        nullable=True, alias="Common Values"
    )


class GeneratedDictDescribed(GeneratedDict):
    description: Series[str] = pa.Field(unique=True, alias="Description")


class DataDictionaryEntry(pa.DataFrameModel):
    source_field: Series[str] = pa.Field(unique=True)
    source_description: Series[str] = pa.Field(unique=True)
    source_type: Series[str]
    common_values: Optional[Series[str]] = pa.Field(nullable=True)
    choices: Optional[Series[str]] = pa.Field(nullable=True)

    @pa.dataframe_check(error="Require exactly one: 'common_values' or 'choices'")
    def one_value_column(cls, df: pa.DataFrameModel):
        cols_present = [
            col for col in ["common_values", "choices"] if col in df.columns
        ]
        return len(cols_present) == 1


class DataDictionaryProcessed(DataDictionaryEntry):
    common_values: Optional[Series[Any]] = pa.Field(nullable=True)
    choices: Optional[Series[dict[str, str]]] = pa.Field(nullable=True)

    @pa.check("common_values")
    def check_list_of_strings(cls, s: Series[Any]) -> Series[bool]:
        return s.apply(
            lambda x: isinstance(x, (list, np.ndarray))
            and all(isinstance(i, str) for i in x)
        )



"Contains all functions that call OpenAI's API."

from __future__ import annotations


class LLMBase:
    def __init__(self, api_key, model=None):  # pragma: no cover
        self.client = None
        self.model = model

    def get_definitions(self, headers, language):  # pragma: no cover
        """
        Get the definitions of the columns in the dataset.
        """
        # subclasses should implement this method
        raise NotImplementedError

    def map_fields(self, source_fields, target_fields):  # pragma: no cover
        """
        Calls an LLM API to generate a draft mapping between two datasets.
        """
        # subclasses should implement this method
        raise NotImplementedError

    def map_values(self, values, language):  # pragma: no cover
        """
        Calls an LLM API to generate a set of value mappings for the fields.
        """
        # subclasses should implement this method
        raise NotImplementedError

    def map_long_table(
        self, single_field_format, descriptions, enums
    ):  # pragma: no cover
        """
        Calls an LLM API to generate a mapping for a long table.
        """
        # subclasses should implement this method
        raise NotImplementedError



from __future__ import annotations

from pathlib import Path
from typing import Literal

import pandas as pd

from adtl import Parser


def parse(
    spec: str | Path | dict[str, str],
    file: str | Path | pd.DataFrame,
    output=None,
    encoding: str = "utf-8-sig",
    include_defs=[],
    save_as: Literal["csv", "parquet", None] = "csv",
    quiet=False,
    verbose=False,
    parallel=False,
):
    """Parse a file according to a specification

    Args:
        spec: Specification file to use
        file: File to parse
        encoding: Encoding of the file
        include_def: Additional definitions to include
        save_as: Save the output as a CSV or parquet file, or don't save (None)
        quiet: Suppress all terminal output
        verbose: Increase verbosity, show overwriting warnings
        parallel: Use parallel processing for parsing. Not recommended for small datasets

    Returns:
        dict[str, pd.DataFrame]: Dictionary of tables parsed into new format
    """
    spec = Parser(
        spec, include_defs=include_defs, quiet=quiet, verbose=verbose, parallel=parallel
    )

    # check for incompatible options
    if spec.header.get("returnUnmatched") and save_as == "parquet":
        raise ValueError("returnUnmatched and parquet options are incompatible")

    # run adtl
    adtl_output = spec.parse(file, encoding=encoding)
    if save_as:
        adtl_output.save(output or spec.name, save_as)
    return {k: pd.DataFrame(v) for k, v in adtl_output.data.items()}



"""
Infer a data dictionary from a dataset.
"""

from __future__ import annotations

import argparse
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from pandera.errors import SchemaError

from .config.config import get_config, setup_config
from .data_dict_schema import GeneratedDict, GeneratedDictDescribed
from .util import (
    DEFAULT_CONFIG,
    check_matches,
    load_data_dict,
    read_data,
)


class DictWriter:
    """
    Class for inferring a data dictionary based on a dataset. Will not store the data,
    only the created data dictionary.

    Use `create_dict()` to create a data dictionary, as the function equivalent
    of the command line `create-dict` script.

    `generate_descriptions()` will use an LLM to generate descriptions for the
    data dictionary, using only the column headers, NOT the data itself.

    Parameters
    ----------
    api_key
        API key corresponsing to the chosen LLM provider/model
    """

    def __init__(self):
        self.config = get_config()
        self.model = self.config._llm

    def _reset_headers_and_validate(self, data_dict: pd.DataFrame) -> pd.DataFrame:
        """
        Reset the headers of a data dictionary to those provided.
        Validate the data dictionary against the generated dictionary schema.


        Returns
        -------
        pd.DataFrame
            Data dictionary with the headers reset
        """

        column_mappings = self.config.column_mappings.model_dump()
        data_dict.rename(columns=column_mappings, inplace=True)

        # Validate the new data dictionary
        try:
            GeneratedDictDescribed.validate(data_dict)
        except SchemaError as e:
            if "'Description' contains duplicate values" in str(e):
                warnings.warn(
                    "Duplicate descriptions found in the data dictionary. "
                    "Before proceeding to mapping and parser generation, each description must be unique."
                )
            else:
                raise e

        return data_dict

    def _load_dict(self, data_dict: pd.DataFrame) -> pd.DataFrame:
        dd = load_data_dict(data_dict, schema=GeneratedDict)
        column_mappings = {
            v: k for k, v in self.config.column_mappings.model_dump().items()
        }
        dd.rename(columns=column_mappings, inplace=True)

        return dd

    def create_dict(self, data: pd.DataFrame | str) -> pd.DataFrame:
        """
        Create a basic data dictionary from a dataset.

        Creates a data dictionary from a dataset, including the field name, field type,
        and common values (defined as occuring more than 25 times in the columns).
        Also creates an empty column for field decriptions, which can either be added by
        hand later, or auto-generated with an LLM using `generate_descriptions()`.

        Parameters
        ----------
        data
            Path to a CSV or XLSX file, or a DataFrame, containing the raw data.

        Returns
        -------
        pd.DataFrame
            Data dictionary containing field names, field types, and common values.
        """

        df = read_data(data, "Data")

        names = df.columns
        types = [str(t) for t in df.dtypes]
        value_opts = {}

        # Get common value thresholds
        max_common_count = self.config.max_common_count
        min_common_freq = self.config.min_common_frequency

        # check the max count isn't > than 30% of the dataset
        calced_max_common_count = min(max_common_count, len(df) * 0.3)
        if calced_max_common_count < max_common_count:
            warnings.warn(
                f"Small Dataset: max_common_count of '{max_common_count}' is"
                f" too high for a dataset with {len(df)} rows.\n"
                f"Reducing to {calced_max_common_count} to avoid data "
                "identification issues.\n"
                "Setting the minimum frequency to 5% of the dataset."
            )
            max_common_count = calced_max_common_count
            min_common_freq = 0.05

        for j, i in enumerate(df.columns):
            values = df[i].value_counts(sort=True)

            # check for lists in the data
            if types[j] == "object" and any(
                "[" in x or "," in x for x in values.index.values if isinstance(x, str)
            ):
                # the values might be lists.
                list_col = df[i].apply(
                    lambda x: (
                        [v.lstrip(" ").rstrip(" ") for v in x.strip("[]").split(",")]
                        if isinstance(x, str)
                        else x
                    )
                )
                flat_col = [item for sublist in list_col.dropna() for item in sublist]
                values = pd.Series(flat_col).value_counts(sort=True)
                types[j] = "list"

            if min_common_freq:
                values = values[values > max(1, len(df) * min_common_freq)]
            value_opts[i] = np.nan
            if not values.empty and len(values) <= max_common_count:
                # drop any values with a frequency of 1
                values = values[values > 1]
                if not values.empty:
                    index_values = list(values.index.values)
                    # Check: only allow if all values are str or bool
                    if all(isinstance(v, (str, bool)) for v in index_values):
                        value_opts[i] = f"{self.config.choice_delimiter} ".join(
                            str(v) for v in index_values
                        )

                    if all(isinstance(v, (bool)) for v in index_values):
                        types[j] = "boolean"

        dd = pd.DataFrame(
            {
                "Field Name": names,
                "Description": [np.nan] * len(names),
                "Field Type": types,
                "Common Values": value_opts.values(),
            }
        )
        dd["Field Type"] = dd["Field Type"].map(
            {
                "object": None,
                "int64": "number",
                "float64": "number",
                "datetime64[ns]": "date",
                "boolean": "bool",
                "list": "list",
            }
        )
        dd["Field Type"] = dd["Field Type"].fillna("string")

        self.data_dictionary = dd

        return dd

    def generate_descriptions(
        self, data_dict: pd.DataFrame | str | None = None
    ) -> pd.DataFrame:
        """
        Generate descriptions for the columns in the dataset.

        Uses an LLM to auto-generate descriptions for a data dictionary based on the
        column headers.

        Parameters
        ----------
        data_dict
            Data dictionary containing the column headers, either as a dataframe or a
            path to the dictionary as a csv/xlsx file. Can be None if the data dict
            has already been created using `create_dict()`.

        Returns
        -------
        pd.DataFrame
            Data dictionary with descriptions added
        """
        if data_dict is None:
            try:
                data_dict = self.data_dictionary
            except AttributeError:
                raise ValueError(
                    "No data dictionary found. Please create a data dictionary first."
                )

        if self.model is None:
            self.config.check_llm_setup()

        df = self._load_dict(data_dict)

        headers = df.source_field

        descriptions = self.model.get_definitions(list(headers), self.config.language)

        descriptions = {d.field_name: d.translation for d in descriptions}
        df_descriptions = pd.DataFrame(
            descriptions.items(), columns=["source_field_gpt", "description"]
        )

        descrip = pd.concat([headers, df_descriptions], axis=1)

        # check ordering is correct even if the return field names aren't quite the same
        # e.g. numbering has been stripped
        assert all(
            descrip.apply(
                lambda x: check_matches(x["source_field_gpt"], [x.source_field]),
                axis=1,
            )
        ), "Field names from the LLM don't match the originals."

        descrip.drop(columns=["source_field_gpt"], inplace=True)

        new_dd = pd.merge(df, descrip, on="source_field")
        new_dd["source_description"] = new_dd["description"]
        new_dd.drop(columns=["description"], inplace=True)

        new_dd = self._reset_headers_and_validate(new_dd)

        return new_dd


def create_dict(data: pd.DataFrame | str) -> pd.DataFrame:
    """
    Create a basic data dictionary from a dataset.

    Creates a data dictionary from a dataset, including the field name, field type,
    and common values (defined as occuring more than 25 times in the columns).
    Also creates an empty column for field decriptions, which can either be added by
    hand later, or auto-generated with an LLM using `generate_descriptions()`.

    Parameters
    ----------
    data
        Path to a CSV or XLSX file, or a DataFrame, containing the raw data.

    Returns
    -------
    pd.DataFrame
        Data dictionary containing field names, field types, and common values.
    """

    dd = DictWriter().create_dict(data)
    return dd


def generate_descriptions(data_dict: pd.DataFrame | str) -> pd.DataFrame:
    """
    Generate descriptions for the columns in the dataset.

    Uses an LLM to auto-generate descriptions for a data dictionary based on the
    column headers.

    Parameters
    ----------
    data_dict
        Data dictionary containing the column headers, either as a dataframe or a path
        to the dictionary as a csv/xlsx file.

    Returns
    -------
    pd.DataFrame
        Data dictionary with descriptions added
    """

    dd = DictWriter().generate_descriptions(data_dict)

    return dd


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Generate a basic data dictionary from a dataset",
        prog="autoparser create-dict",
    )
    parser.add_argument("data", help="Data to create dictionary from")
    parser.add_argument(
        "-d",
        "--descriptions",
        help="Use an LLM to generate descriptions from file headers",
        action="store_true",
    )
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )
    parser.add_argument(
        "-o", "--output", help="Name to use for output files", default="datadict"
    )

    args = parser.parse_args(argv)

    setup_config(args.config or DEFAULT_CONFIG)

    config = get_config()
    if args.descriptions and not config._llm:
        config.check_llm_setup()

    df = create_dict(args.data)
    if args.descriptions:
        df = generate_descriptions(df)

    df.to_csv(f"{args.output}.csv", index=False)


if __name__ == "__main__":
    main()



"""Functions which can be applied to source fields, allowing extensibility"""

from __future__ import annotations

from datetime import date, datetime, timedelta
from math import floor
from typing import Any, Literal

from dateutil.relativedelta import relativedelta

try:
    import zoneinfo
except ImportError:  # pragma: no cover
    from backports import zoneinfo  # noqa

import re
import warnings

import pint


class AdtlTransformationWarning(UserWarning):
    pass


def isNotNull(value: str | None) -> bool:
    "Returns whether value is not null or an empty string"
    return value not in [None, ""]


def textIfNotNull(field: str, return_val: Any) -> Any:
    "Returns a default value if field is not null"
    return return_val if field not in [None, ""] else None


def wordSubstituteSet(value: str, *params) -> list[str]:
    """
    For a value that can have multiple words, use substitutions from params.

    Args:
        value: String containing a list of words that should be substituted
        params: List of 2-tuples, in the form
            [(w1, s1), (w2, s2), ... (w_n, s_n)]
            where w1 is replaced by s1, w2 is replaced by s2.

            Word matches are regular expressions, delimited by the `\b` word
            boundary delimiter so can have arbitrary regular expressions to
            match. Any match of regex w_n will use substitute s_n. Case is
            ignored when matching.

    Returns:
        List of words after finding matches and substituting.
        Duplicate words are only represented once.
    """
    out = []
    for i in params:
        if not isinstance(i, (tuple, list)) or not all(isinstance(s, str) for s in i):
            raise ValueError("wordSubstituteSet: params item not a tuple or list")
        sub_map = dict(params)
        for match, subst in sub_map.items():
            if re.search(r"\b" + match + r"\b", value, re.IGNORECASE):
                out.append(subst)
    if not out and (value not in [None, ""]):
        warnings.warn(
            f"No matches found for: '{value}'",
            AdtlTransformationWarning,
            stacklevel=2,
        )
    return sorted(set(out)) if out else None


def getFloat(value: str, set_decimal: str | None = None, separator: str | None = None):
    """Returns value transformed into a float.

    Args:
        value: Value to be transformed to float
        set_decimal: optional, set if decimal separator is not a
            full stop or period (.)
        separator: optional, set to the character used for separating
            thousands (such as `,`).
    """

    if not value:
        return None

    if '"' in value or " " in value:
        value = value.strip('"').replace(" ", "")

    if set_decimal:
        # handle comma decimal separator
        # partition always splits on last instance so copes if decimal == separator
        value_int, _, fraction = value.partition(set_decimal)
        value = value_int + "." + fraction

    if separator:
        if separator in value and separator != ".":
            value = value.replace(separator, "")
        elif separator in value_int:
            value = value_int.replace(separator, "") + "." + fraction

    values = [float(d) for d in re.findall(r"[-+]?\d*\.?\d+", value)]
    value = values[0] if len(values) == 1 else value

    try:
        return float(value)
    except ValueError:
        return value if value != "" else None


def Percentage(value: float):
    "transform a decimal into a percentage"
    try:
        value = float(value)
    except (ValueError, TypeError):
        return value

    if value > 1:
        return value

    return value * 100


def yearsElapsed(
    birthdate: str,
    currentdate: str,
    epoch: float,
    bd_format: str = "%Y-%m-%d",
    cd_format: str = "%Y-%m-%d",
):
    """
    Returns the number of years elapsed between two dates, useful for calculating ages

    Args:
        birthdate: Start date of duration
        currentdate: End date of duration
        epoch: Epoch year after which dates will be converted to the last century.
            As an example, if epoch is 2022, then the date 1/1/23 will be converted
            to the January 1, 1923.
        bd_format: Date format for *birthdate* specified using :manpage:`strftime(3)`
            conventions. Defaults to ISO format ("%Y-%m-%d")
        cd_format: Date format for *currentdate* specified using :manpage:`strftime(3)`
            conventions. Defaults to ISO format ("%Y-%m-%d")

    Returns:
        int | None: Number of years elapsed or None if invalid dates were encountered
    """

    if birthdate in [None, ""] or currentdate in [None, ""]:
        return None

    bd = correctOldDate(birthdate, epoch, bd_format, return_datetime=True)

    if bd is None:
        return None

    cd = datetime.strptime(currentdate, cd_format)

    try:
        days = cd - bd
        return pint.Quantity(days.days, "days").to("years").m
    except ValueError:
        warnings.warn(
            f"Failed calculation yearsElapsed: {birthdate}, {currentdate}",
            AdtlTransformationWarning,
            stacklevel=2,
        )


def durationDays(startdate: str, currentdate: str) -> int:
    """
    Returns the number of days between two dates.
    Preferable to Y-M-D elapsed, as month length is ambiguous -
    can be anywhere between 28-31 days.
    """
    if startdate in [None, ""] or currentdate in [None, ""]:
        return None

    bd = datetime.strptime(startdate, "%Y-%m-%d")
    cd = datetime.strptime(currentdate, "%Y-%m-%d")

    days = cd - bd
    return days.days


def startDate(enddate: str, duration: str) -> str:
    """
    Returns the start date in ISO format, given the end date and the duration.
    """
    if enddate in [None, ""] or duration in [None, ""]:
        return None

    ed = datetime.strptime(enddate, "%Y-%m-%d")

    sd = ed - timedelta(days=float(duration))

    return sd.strftime("%Y-%m-%d")


def endDate(startdate: str, duration: str, format="%Y-%m-%d"):
    """
    Returns the end date in ISO format, given the start date and the duration.

    Args:
        startdate: Start date
        duration: Duration in days
        format: :manpage:`strftime(3)` format that dates are in

    Returns:
        End date in the specified format
    """
    if startdate in [None, ""] or duration in [None, ""]:
        return None

    sd = datetime.strptime(startdate, format)

    duration = float(duration)

    ed = sd + timedelta(days=duration)

    return ed.strftime("%Y-%m-%d")


def makeDate(year: str, month: str, day: str) -> str:
    "Returns a date from components specified as year, month and day"
    if year in [None, ""] or month in [None, ""] or day in [None, ""]:
        return None
    try:
        year, month, day = int(year), int(month), int(day)
    except ValueError:
        warnings.warn(
            f"Could not construct date from: year={year}, month={month}, day={day}",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None
    try:
        return date(year, month, day).isoformat()
    except ValueError:
        warnings.warn(
            f"Could not construct date from: year={year}, month={month}, day={day}",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None


def makeDateTimeFromSeconds(
    date: str, time_seconds: int, date_format: str, timezone: str
) -> datetime:
    """Returns a datetime from date and time specified in
    elapsed seconds since the beginning of the day

    Args:
        date: Date to be converted
        time_seconds: Elapsed time in seconds within that day (0 - 86399)
        date_format: Date format in :manpage:`strftime(3)` format
        timezone: Timezone to use, specified in tzdata format

    Returns:
        datetime.datetime: A timezone aware datetime object
    """
    if date == "":
        return None
    try:
        t = datetime.strptime(date, date_format).replace(
            tzinfo=zoneinfo.ZoneInfo(timezone)
        )
    except ValueError:
        warnings.warn(
            f"Could not convert date {date!r} from date format {date_format!r}",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None
    if time_seconds == "":
        return t.date().isoformat()  # return date only
    time_seconds = int(time_seconds)
    hour = time_seconds // 3600
    minute = (time_seconds % 3600) // 60
    return t.replace(hour=hour, minute=minute).isoformat()


def makeDateTime(
    date: str, time_24hr: str, date_format: str, timezone: str
) -> datetime:
    """Returns a combined date and time

    Args:
        date: Date to be converted
        time_24hr: Time specified in HH:MM format
        date_format: Date format in :manpage:`strftime(3)` format
        timezone: Timezone to use, specified in tzdata format

    Returns:
        datetime.datetime: A timezone aware datetime object
    """
    if date == "":
        return None

    try:
        dt = datetime.strptime(date, date_format).replace(
            tzinfo=zoneinfo.ZoneInfo(timezone)
        )
    except ValueError:
        warnings.warn(
            f"Could not convert date {date!r} from date format {date_format!r}",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None

    if time_24hr == "":
        return dt.date().isoformat()  # return date only

    tm = datetime.strptime(time_24hr, "%H:%M").time()

    return datetime.combine(dt, tm, tzinfo=zoneinfo.ZoneInfo(timezone)).isoformat()


def splitDate(
    date: str,
    option: Literal["year", "month", "day"],
    epoch: float,
    format: str = "%Y-%m-%d",
):
    "Splits a date into year, month, day"

    if date in [None, ""]:
        return None

    sd = correctOldDate(date, epoch, format, return_datetime=True)

    if sd is None:
        return None

    if option == "year":
        return sd.year
    elif option == "month":
        return sd.month
    elif option == "day":
        return sd.day
    else:
        warnings.warn(
            f"Invalid option {option!r} for splitDate",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None


def startYear(
    duration: str | float,
    currentdate: list | str,
    epoch: float,
    dateformat: str = "%Y-%m-%d",
    duration_type: Literal["years", "months", "days"] = "years",
    provide_month_day: bool | list = False,
) -> int | float:
    """
    Use to calculate year e.g. of birth from date (e.g. current date) and
    duration (e.g. age)

    The date can be provided as a list of possible dates (if a hierarchy needs
    searching through)

    Args:
        duration: Duration value
        currentdate: Date to offset duration from
        epoch: Epoch year to use for conversion of two digit years. Any dates
            after the epoch are converted to the last century
        dateformat: Date format that currentdate is in
        duration_type: One of 'years', 'months' or 'days'
        provide_month_day: If currentdate is only year, and this is specified
            as a tuple of (month, day), uses these to construct the date

    Returns:
        Starting year, offset by duration
    """
    if isinstance(currentdate, list):
        # find the first non nan instance, else return None
        currentdate = next((s for s in currentdate if s), None)

    if currentdate in [None, ""] or duration in [None, ""]:
        return None

    if provide_month_day:
        cd = makeDate(currentdate, provide_month_day[0], provide_month_day[1])
        cd = datetime.strptime(cd, "%Y-%m-%d")
    else:
        cd = correctOldDate(currentdate, epoch, dateformat, return_datetime=True)

    if cd is None:
        return None

    if duration_type == "years":
        return cd.year - floor(float(duration))

    elif duration_type == "months":
        new_date = cd - relativedelta(months=floor(float(duration)))  # rounds down
        return new_date.year

    elif duration_type == "days":
        new_date = cd - timedelta(days=float(duration))
        return new_date.year


def startMonth(
    duration: str | float,
    currentdate: list | str,
    epoch: float,
    dateformat: str = "%Y-%m-%d",
    duration_type: Literal["years", "months", "days"] = "years",
    provide_month_day: bool | list = False,
):
    """
    Use to calculate month e.g. of birth from date (e.g. current date) and
    duration (e.g. age), parameter descriptions are same as
    :meth:`adtl.transformations.startYear`, except this function
    returns the month component
    """
    if isinstance(currentdate, list):
        # find the first non nan instance, else return None
        currentdate = next((s for s in currentdate if s), None)

    if currentdate in [None, ""] or duration in [None, ""]:
        return None

    if provide_month_day:
        cd = makeDate(currentdate, provide_month_day[0], provide_month_day[1])
        cd = datetime.strptime(cd, "%Y-%m-%d")
    else:
        cd = correctOldDate(currentdate, epoch, dateformat, return_datetime=True)

    if cd is None:
        return None

    if duration_type == "months":
        new_date = cd - relativedelta(months=floor(float(duration)))
        return new_date.month

    elif duration_type == "days":
        new_date = cd - timedelta(days=float(duration))
        return new_date.month


def correctOldDate(date: str, epoch: float, format: str, return_datetime: bool = False):
    """
    Fixes dates so that they are the correct century for when the year
    is not fully specified. The time module converts 2 digit dates by
    mapping values 69-99 to 1969-1999, and values 0-68 are mapped to
    2000-2068. This doesn't work for e.g. birthdates here where they are
    frequently below the cutoff.

    Switches the pivot point to that set by epoch so that epoch+ converts to 19xx.

    Only use for birth dates to avoid unintentional conversion for recent
    dates.

    Args:
        date: Date to convert
        epoch: Epoch as year
        format: :manpage:`strftime(3)` format that date is in
        return_datetime: Whether to return date in a datetime.datetime format
            (when True), or a string (when False, default)

    Returns:
        str | datetime.datetime: Fixed date, return type depends on return_datetime
    """

    if date in [None, ""]:
        return None

    try:
        cd = datetime.strptime(date, format)
    except ValueError:
        warnings.warn(
            f"Could not convert date {date!r} from date format {format!r}",
            AdtlTransformationWarning,
            stacklevel=2,
        )
        return None

    if cd.year >= epoch and "y" in format:
        cd = cd.replace(year=cd.year - 100)

    if return_datetime:
        return cd
    else:
        return cd.strftime("%Y-%m-%d")



from __future__ import annotations

import warnings
from functools import cached_property
from typing import Generator

import numpy as np
import pandas as pd

from ..util import check_matches
from .base_mapper import BaseMapper


class WideMapper(BaseMapper):
    """
    Class for creating an intermediate mapping file linking the data dictionary to
    a wide schema fields and values.
    """

    @cached_property
    def target_fields(self) -> list[str]:
        """Returns a list of fields in the target schema"""
        return list(self.schema_fields.keys())

    @cached_property
    def target_types(self) -> dict[str, list[str]]:
        """Returns the field types of the target schema"""
        return {
            f: self.schema_fields[f].get("type", ["string", "null"])
            for f in self.target_fields
        }

    @cached_property
    def target_values(self) -> pd.Series:
        """Returns the enum values or boolean options for the target schema"""

        def _value_options(f):
            if "boolean" in self.target_types[f]:
                return ["True", "False", "None"]
            elif "string" in self.target_types[f]:
                return self.schema_fields[f].get("enum", np.nan)
            elif "array" in self.target_types[f]:
                return self.schema_fields[f].get("items", {}).get("enum", np.nan)
            else:
                return np.nan

        target_vals = pd.Series(
            {f: _value_options(f) for f in self.target_fields},
            self.target_fields,
            name="target_values",
        )
        target_vals.index.name = "target_field"
        return target_vals

    def _iter_value_tuples(self) -> Generator[tuple[str, list[str], list[str]]]:
        for f in self.target_fields:
            s = self.common_values_mapped.get(f)
            t = self.target_values[f]
            if s is not None and t is not None:
                yield (f, s, t)

    def match_fields_to_schema(self) -> pd.DataFrame:
        """
        Use the LLM to match the target (schema) fields to the descriptions of the
        source data fields from the data dictionary.
        """

        # english translated descriptions rather than names.
        source_fields = list(self.data_dictionary.source_description)

        mappings = self.model.map_fields(source_fields, self.target_fields)

        mapping_dict = pd.DataFrame(
            {
                "target_field": [f.target_field for f in mappings.targets_descriptions],
                "source_description": [
                    f.source_description for f in mappings.targets_descriptions
                ],
            }
        )

        df_merged = pd.merge(
            mapping_dict,
            self.data_dictionary,
            how="left",
            on="source_description",
        ).drop_duplicates(subset="target_field")
        df_merged.set_index("target_field", inplace=True, drop=True)

        # Check to see if any fields with mapped descriptions are missing after merge
        missed_merge = df_merged[
            (df_merged["source_description"].notna())
            & (df_merged["source_field"].isna())
        ]

        if not missed_merge.empty:
            descriptions_list = (
                self.data_dictionary["source_description"].dropna().tolist()
            )
            df_merged.loc[missed_merge.index, "source_description"] = missed_merge[
                "source_description"
            ].apply(lambda x: check_matches(x, descriptions_list))

            df_merged = (
                df_merged["source_description"]
                .reset_index()
                .merge(self.data_dictionary, how="left")
                .drop_duplicates(subset="target_field")
                .set_index("target_field")
            )

        self.mapped_fields = df_merged.source_field
        self.filtered_data_dict = df_merged
        return df_merged

    def create_mapping(self, save=True, file_name="mapping_file") -> pd.DataFrame:
        """
        Creates an intermediate mapping dataframe linking the data dictionary to schema
        fields. The index contains the target (schema) field names, and the columns are:
        * source_description
        * source_field
        * common_values OR choices (depending on the data dictionary)
        * target_values
        * value_mapping

        Raises a warning if any fields are present in the schema where a
        corresponding source field in the data dictionary has not been found.

        Parameters
        ----------
        save
            Whether to save the mapping to a CSV file. If True, lists and dicts are
            converted to strings before saving.
        file_name
            The name to use for the CSV file
        """

        mapping_dict = self.match_fields_to_schema()
        mapped_vals = self.match_values_to_schema()

        mapping_dict.drop(columns=["source_type"], inplace=True)

        # reindex to add in any schema fields that weren't returned by the LLM
        mapping_dict = mapping_dict.reindex(self.target_fields)

        mapping_dict = pd.concat(
            [mapping_dict, self.target_values, mapped_vals], axis=1
        )

        unmapped = mapping_dict[mapping_dict["source_field"].isna()].index
        if any(unmapped):
            warnings.warn(
                f"The following schema fields have not been mapped: {list(unmapped)}",
                UserWarning,
            )

        return self.post_process_mapping(mapping_dict, save=save, file_name=file_name)



from __future__ import annotations

import logging
import warnings
from typing import Callable

import fastjsonschema

logger = logging.getLogger(__name__)


def convert_to_schema_type(value, target_type: str | list[str]):
    """
    Convert value to the target type specified in a JSON schema.
    """
    type_casters = {
        "string": str,
        "integer": int,
        "number": float,
    }

    if isinstance(target_type, str):
        target_type = [target_type]

    for tt in target_type:
        if tt in type_casters:
            try:
                return type_casters[tt](value)
            except (ValueError, TypeError):
                if tt == "integer":
                    # Special case for converting float to integer with rounding
                    try:
                        return int(round(float(value)))
                    except (ValueError, TypeError):
                        logger.debug(f"Could not convert value {value} to integer")
                        return value

                logger.debug(f"Could not convert value {value} to type {tt}")

    return value


def expand_schema(
    schema: dict, discriminator: str
) -> tuple[Callable | dict[str, Callable], bool]:
    """
    For long table schemas with a set of 'oneOf' options, this function expands the schema
    into a dictionary of validators keyed by the discriminator value for faster validation
    and more useful error messages.

    Parameters
    ----------
    schema : dict
        The JSON schema to expand.
    discriminator : str
        The column in the target table that holds the discriminator/variable/attribute/name values.
    """
    per_attribute_schemas = {}

    general_properties = schema.get("properties", {})
    general_required = schema.get("required", [])

    # if oneOf doesn't hold properties, we can skip the expansion
    oneof_options = any("properties" in opt for opt in schema.get("oneOf", []))

    if oneof_options:
        for subschema in schema.get("oneOf", []):
            props = subschema.get("properties", {})
            required = subschema.get("required", [])
            # Figure out which attributes this subschema applies to
            attr = props.get(discriminator, {})
            attr_keys = []
            if "const" in attr:
                attr_keys = [attr["const"]]
            elif "enum" in attr:
                attr_keys = attr["enum"]
            else:
                warnings.warn(
                    f"Found long schema item '{attr}' with neither 'const' nor 'enum'.\n"
                    "Falling back to unexpanded schema mode"
                )
                per_attribute_schemas = []
                break
            # Build the minimal schema for this attribute
            per_attr_schema = {
                "type": "object",
                "properties": general_properties | props,
                "required": general_required + required,
                "additionalProperties": schema.get("additionalProperties", False),
            }
            for key in attr_keys:
                per_attribute_schemas[key] = per_attr_schema

        if len(per_attribute_schemas) != 0:
            return {
                k: fastjsonschema.compile(v) for k, v in per_attribute_schemas.items()
            }, True
    return fastjsonschema.compile(schema), False



"Contains all functions that call OpenAI's API."

from __future__ import annotations

from openai import OpenAI
from pydantic import create_model

from .base_llm import LLMBase
from .data_structures import ColumnDescriptionRequest, MappingRequest, ValuesRequest


class OpenAILanguageModel(LLMBase):
    def __init__(self, api_key, model: str = "gpt-4o-mini"):
        self.client = OpenAI(api_key=api_key)
        self.model = model

        if self.model not in self.valid_models():
            raise ValueError(
                f"Unsupported OpenAI model. Must be one of {self.valid_models}."
            )

    @classmethod
    def valid_models(cls):
        return ["gpt-4o-mini", "gpt-4o", "o1", "o3-mini"]

    def get_definitions(self, headers: list[str], language: str) -> dict[str, str]:
        """
        Get the definitions of the columns in the dataset.
        """
        completion = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "The following is a list of headers from a data file in "
                        f"{language}, some containing shortened words or abbreviations. "  # noqa
                        "Translate them to english. "
                        "Return a list of (original header, translation) pairs, using the given structure."  # noqa
                    ),
                },
                {"role": "user", "content": f"{headers}"},
            ],
            response_format=ColumnDescriptionRequest,
        )
        descriptions = completion.choices[0].message.parsed.field_descriptions

        return descriptions

    def map_fields(
        self, source_fields: list[str], target_fields: list[str]
    ) -> MappingRequest:
        """
        Calls the OpenAI API to generate a draft mapping between two datasets.
        """
        field_mapping = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "You will be given two lists of phrases, one is the headers "
                        "for a target data file, and the other a set of descriptions "
                        "for columns of source data. "
                        "Match each target header to the best matching source "
                        "description, but match a header to None if a good match does "
                        "not exist. "
                        "Return the matched target headers and source descriptions using the provided structure."  # noqa
                    ),
                },
                {
                    "role": "user",
                    "content": (
                        f"These are the target headers: {target_fields}\n"
                        f"These are the source descriptions: {source_fields}"
                    ),
                },
            ],
            response_format=MappingRequest,
        )
        mappings = field_mapping.choices[0].message.parsed

        return mappings

    def map_values(
        self, values: list[tuple[str, set[str], list[str | None] | None]], language: str
    ) -> ValuesRequest:
        """
        Calls the OpenAI API to generate a set of value mappings for the fields.
        """
        value_mapping = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "You will be given a list of tuples, where each tuple contains "
                        "three sets of string values. "
                        "The first set contains field names for a dataset."
                        "The second set contains values from a source dataset in "
                        f"{language}, and the third set contains target values for an "
                        "english-language transformed dataset. "
                        "Match all the values in the second set to the appropriate "
                        "values in the third set. "
                        "Return a list of dictionaries, where each dictionary contains "
                        "the field name as a key, and a dictionary containing "
                        "source values as keys, and the target text as values, "
                        "as the values. For example, the result should look like this: "
                        "[{'field_name_1': {'source_value_a': 'target_value_a', "
                        "'source_value_b': 'target_value_b'}, 'field_name_2':{...}]"
                        "using the provided structure."
                    ),
                },
                {
                    "role": "user",
                    "content": (
                        f"These are the field, source, target value sets: {values}"
                    ),
                },
            ],
            response_format=ValuesRequest,
        )
        mappings = value_mapping.choices[0].message.parsed

        return mappings

    def map_long_table(self, single_field_format, descriptions, enums):
        """
        Calls the OpenAI API to generate a mapping for a long table.
        """

        LongTableRequest = create_model(
            "LongTableRequest", long_table=(list[single_field_format], ...)
        )

        system_msg_descriptive = """
            You are an expert at structured data extraction.
            For each source column description provided, match it to the best variable name
            from the given list of enums. If no good match exists, use null.

            Return output as a JSON array of objects, each with fields:
            - source_description (string)
            - variable_name (string or null)
            - value_col (one of "value_bool", "value_num", "value" or null)
            - phase (one of "presentation", "outcome", or null)
            - attribute_unit (string or null)

            Example input:

            Columns descriptions: ["Age", "Symptoms: Cough", "Admission Date"]
            Variable names: ["age", "cough", "aids_hiv"]

            Example output:
            [
            {"source_description": "Age", "variable_name": "age", "value_col": "value_num", "phase": "presentation", "attribute_unit": "years"},
            {"source_description": "Symptoms: Cough", "variable_name": "cough", "value_col": "value_bool", "phase": "presentation", "attribute_unit": ""},
            {"source_description": "Admission Date", "variable_name": null, "value_col": "", "phase": "", "attribute_unit": ""}
            ]
            """

        long_table_mapping = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": system_msg_descriptive,
                },
                {
                    "role": "user",
                    "content": (
                        f"Columns descriptions: {', '.join(descriptions)}"
                        f"Variable names: {', '.join(enums)}"
                    ),
                },
            ],
            response_format=LongTableRequest,
        )
        mappings = long_table_mapping.choices[0].message.parsed

        return mappings
